﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework;

namespace Collisions2D
{
    class Character : DrawableGameComponent
    {

        private SpriteBatch spriteBatch;
        private Matrix currentTransformation;

        // Indice dell'animazione corrente
        private int currentAnimationIndex = 0;
        // Array contenente tutte le animazioni del Character
        public Animation[] Animations { get; set; }
        // Posizione corrente del Character
        public Vector2 Position { get; set; }
        // Texture che useremo per disegnare i rettangoli che 
        // contengono la figura
        private Texture2D violetPixel, bluePixel;

        // Costruttori
        public Character(Game game, Vector2 InitialPosition) : base(game) { this.Position = InitialPosition; }
        public Character(Game game) : this(game, new Vector2(0, 0)) { }

        protected override void LoadContent()
        {
            base.LoadContent();

            // Otteniamo lo spriteBatch del gioco
            spriteBatch = (SpriteBatch)Game.Services.GetService(typeof(SpriteBatch));

            // Inizializziamo le texture per disegnare i rettangoli
            // contengono la figura
            violetPixel = ((Collisions2DGame) this.Game).CreatePixel(Color.DarkViolet, 32);
            bluePixel = ((Collisions2DGame) this.Game).CreatePixel(Color.Blue, 32);
        }

        public override void Update(GameTime gameTime)
        {
            base.Update(gameTime);

            // Ad ogni update aggiorniamo la matrice di trasformazione.
            // Essa contiene tutte le informazioni su centro di rotazione,
            // angolo di rotazione, ridimensionamento e posizione
            this.currentTransformation = 
                Matrix.CreateTranslation(new Vector3(-this.CurrentAnimation.RotationPoint, 0))
                * Matrix.CreateScale(this.CurrentAnimation.Scale)
                * Matrix.CreateRotationZ((float)this.CurrentAnimation.CurrentRotation)
                * Matrix.CreateTranslation(new Vector3(this.Position, 0));
            
            // Aggiorna le variabili dell'animazione
            while (!this.CurrentAnimation.Update(gameTime))
            {
                // L'animazione è finita, passiamo alla successiva
                Animation oldAnimation = CurrentAnimation;
                CurrentAnimation = Animations[(currentAnimationIndex + 1) % Animations.Length];
                CurrentAnimation.Start(gameTime);
                CurrentAnimation.InitalAngle = oldAnimation.CurrentRotation;
            }
        }

        public override void Draw(GameTime gameTime)
        {
            base.Draw(gameTime);
#if !DRAW_JUST_COLLISIONS
            // disegniamo il rettangolo che contiene la figura prima
            // delle trasformazioni
            spriteBatch.Draw(bluePixel,
                this.Position, this.CurrentAnimation.CurrentRectangle,
                Color.White, (float)this.CurrentAnimation.CurrentRotation,
                this.CurrentAnimation.RotationPoint,
                this.CurrentAnimation.Scale, SpriteEffects.None, 0);

            // disegniamo il rettangolo che contiene la figura dopo le 
            // trasformazioni, questo è quello che viene utilizzato
            // come primo indicatore di una possibile collisione
            spriteBatch.Draw(violetPixel, this.GetBoundingRectangle(), Color.White);
#endif
            // disegniamo il character con tutte le informazioni
            // che ci vengono da this.CurrentAnimation, eccetto
            // la posizione corrente
            spriteBatch.Draw(this.CurrentAnimation.Frames,
                this.Position, this.CurrentAnimation.CurrentRectangle,
                Color.White, (float)this.CurrentAnimation.CurrentRotation,
                this.CurrentAnimation.RotationPoint,
                this.CurrentAnimation.Scale, SpriteEffects.None, 0);    

        }

        // Calcoliamo il rettangolo che contiene la figura 
        // dopo le trasformazioni
        public Rectangle GetBoundingRectangle() {
            // Rettangolo prima delle trasformazioni
            Rectangle dest = new Rectangle(0, 0, this.CurrentAnimation.FrameWidth, this.CurrentAnimation.Frames.Height);
            
            // Prendiamo la matrice di trasformazione
            Matrix transformation = this.currentTransformation;

            // Creiamo i vettori dei punti dei 4 angoli del rettangolo
            Vector2 ul = new Vector2(dest.Left, dest.Top), 
                ur = new Vector2(dest.Right, dest.Top),
                dl = new Vector2(dest.Left, dest.Bottom), 
                dr = new Vector2(dest.Right, dest.Bottom);

            // Applichiamo a tutti gli angoli la trasformazione
            ul = Vector2.Transform(ul, transformation);
            ur = Vector2.Transform(ur, transformation);
            dl = Vector2.Transform(dl, transformation);
            dr = Vector2.Transform(dr, transformation);

            // Prendiamo il rettangolo spostato, scalato e ruotato ottenuto 
            // e ne calcoliamo il rettangolo che a sua volta lo contiene.
            // Questo ci serve perché dobbiamo lavorare con un rettangolo
            // con i lati paralleli agli assi (ovvero allo schermo).

            // Calcoliamo l'angolo in alto a destra e in basso a sinistra
            // prendendo le coordinate minime e massime dei punti traslati
            Vector2 ulb = Vector2.Min(Vector2.Min(ul, ur), Vector2.Min(dl, dr));
            Vector2 drb = Vector2.Max(Vector2.Max(ul, ur), Vector2.Max(dl, dr));

            // Creiamo e restituiamo il rettangolo ottenuto
            return new Rectangle((int) Math.Round(ulb.X), (int) Math.Round(ulb.Y), 
                (int) Math.Round(drb.X - ulb.X), (int) Math.Round(drb.Y - ulb.Y));
        }

        // Funzione statica che dati due rettangoli, ne trova l'intersezione (se c'è)
        // e la restituisce
        public static Rectangle? GetRectanglesIntersection(Rectangle rect1, Rectangle rect2)
        {
            if (rect1.Intersects(rect2)) {
                Vector2 ul = Vector2.Max(new Vector2(rect1.Left,rect1.Top), new Vector2(rect2.Left, rect2.Top));
                Vector2 dr = Vector2.Min(new Vector2(rect1.Right, rect1.Bottom), new Vector2(rect2.Right, rect2.Bottom));
                return new Rectangle((int) Math.Round(ul.X), (int) Math.Round(ul.Y), 
                    (int) Math.Round(dr.X - ul.X), (int) Math.Round(dr.Y - ul.Y));
            } 
            else { return null; }
        }

        // Funzione che restituisce l'intersezione tra i rettangoli che contengono
        // le due figure da confrontare
        public static Rectangle? GetBoundingRectanglesIntersection(Character obj1, Character obj2) {
            return GetRectanglesIntersection(obj1.GetBoundingRectangle(), obj2.GetBoundingRectangle());
        }

        // CheckCollision che ignora il rettangolo di intersezione
        public static bool CheckCollision(Character obj1, Character obj2) {
            Rectangle ignore;
            return CheckCollision(obj1, obj2, out ignore);
        }

        // Funzione statica che controlla se due Character stanno collidendo
        public static bool CheckCollision(Character obj1, Character obj2, out Rectangle BoundingRectangleIntersection) { 
            BoundingRectangleIntersection = Rectangle.Empty;

            // Prendiamo il rettangolo di intersezione tra i due rettangoli
            // che contengono le figure di cui si sta cercando l'intersezione
            Rectangle? getIntersectionResult = Character.GetBoundingRectanglesIntersection(obj1, obj2);

            // Se neppure i rettangoli delle figure si intersecano sicuramente
            // non c'è collisione
            if (getIntersectionResult == null)
            {
                return false;
            }

            // C'è intersezione tra i rettangoli che contengono le figure
            Rectangle intersection = (Rectangle)getIntersectionResult;
            BoundingRectangleIntersection = intersection;

            // Eseguiamo una serie di operazioni sia per il primo Character
            // che per il secondo

            // Invertiamo la matrice di trasformazione
            Matrix invertedTransform1 = Matrix.Invert(obj1.currentTransformation);
            // Prendiamo il primo punto dell'intersezione dei rettangoli
            // e lo calcoliamo nel sistema di riferimento del Character
            // con cui stiamo lavorando
            Vector2 rowPos1 = Vector2.Transform(new Vector2(intersection.X, intersection.Y), invertedTransform1);
            // Calcoliamo i vettori unitari X e Y (ovvero verso destra e verso il basso)
            // nel sistema di riferimento del Character
            Vector2 xUnit1 = Vector2.TransformNormal(Vector2.UnitX, invertedTransform1);
            Vector2 yUnit1 = Vector2.TransformNormal(Vector2.UnitY, invertedTransform1);
            // Memorizziamo in variabili locali alcune informazioni sul Character
            // in modo da avere le massime prestazioni durante le iterazioni sotto
            int textureWidth1 = obj1.CurrentAnimation.Frames.Width,
                frameLeft1 = obj1.CurrentAnimation.CurrentRectangle.Left,
                frameTop1 = obj1.CurrentAnimation.CurrentRectangle.Top,
                frameWidth1 = obj1.Width,
                frameHeight1 = obj1.Height;

            // Replichiamo tutto quanto per l'altro Character
            Matrix invertedTransform2 = Matrix.Invert(obj2.currentTransformation);
            Vector2 rowPos2 = Vector2.Transform(new Vector2(intersection.X, intersection.Y), invertedTransform2);
            Vector2 xUnit2 = Vector2.TransformNormal(Vector2.UnitX, invertedTransform2);
            Vector2 yUnit2 = Vector2.TransformNormal(Vector2.UnitY, invertedTransform2);
            int textureWidth2 = obj2.CurrentAnimation.Frames.Width,
                frameLeft2 = obj2.CurrentAnimation.CurrentRectangle.Left,
                frameTop2 = obj2.CurrentAnimation.CurrentRectangle.Top,
                frameWidth2 = obj2.Width,
                frameHeight2 = obj2.Height;


            // Cicliamo tutti i punti dell'intersezione dei rettangoli
            // alla ricerca di due punti non completamente trasparenti
            // dei due Character che si sovrappongano
            for (int row = 0; row < intersection.Height; row++)
            {
                // Riportiamo la coordinata X all'inizio
                // della riga del rettangolo (sempre nei
                // sistemi di riferimento dei Character)
                Vector2 pos1 = rowPos1;
                Vector2 pos2 = rowPos2;

                for (int col = 0; col < intersection.Width; col++)
                {
                    // Calcoliamo il punto corrente in cui ci troviamo
                    // dal punto di vista dei due Character
                    Point pt1 = new Point((int) Math.Round(pos1.X), (int) Math.Round(pos1.Y));
                    Point pt2 = new Point((int) Math.Round(pos2.X), (int) Math.Round(pos2.Y));

                    // Se il punto cade all'interno dell'area della figura
                    // di entrambi i Character
                    if (0 <= pt1.X && pt1.X < frameWidth1 &&
                        0 <= pt1.Y && pt1.Y < frameHeight1 &&
                        0 <= pt2.X && pt2.X < frameWidth2 &&
                        0 <= pt2.Y && pt2.Y < frameHeight2)
                    {
                        // Controlliamo che l'n-esimo pixel non sia 
                        // trasparente in entrambe le figure.
                        // Per fare questo ci serviamo della maschera
                        // di bit creata in precedenza (si veda 
                        // Animation.BuildBitmask).
                        // La maschera di bit è unidimensionale perciò
                        // calcoliamo la posizione tramite la formula:
                        // CoordinataX + LarghezzaDiUnaRiga * CoordinataY
                        if (obj1.CurrentAnimation.BitMask[frameLeft1 + pt1.X + 
                            (frameTop1 + pt1.Y) * textureWidth1] &&
                            obj2.CurrentAnimation.BitMask[frameLeft2 + pt2.X + 
                            (frameTop2 + pt2.Y) * textureWidth2])
                            return true;

                    }

                    // Avanziamo verso destra (di una colonna) in entrambi i Character
                    pos1 += xUnit1;
                    pos2 += xUnit2;
                }

                // Avanziamo verso il basso (di una riga) in entrambi i Character
                rowPos1 += yUnit1;
                rowPos2 += yUnit2;
            }

            return false;
        }

        public Animation CurrentAnimation
        {
            get { return Animations[currentAnimationIndex]; }
            set
            {
                currentAnimationIndex = Array.IndexOf(Animations, value);
                if (currentAnimationIndex == -1) throw new IndexOutOfRangeException();
            }

        }

        public int Width { get { return this.CurrentAnimation.FrameWidth; } }
        public int Height { get { return this.CurrentAnimation.Frames.Height; } }

    }

}
