using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;

namespace Collisions2D
{
    public class Collisions2DGame : Microsoft.Xna.Framework.Game
    {
        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;

        // Immagine contenente i frame dell'animazione
        Texture2D framesStellina;

        // Stato del tasto sinistro del mouse dell'ultimo
        // Update
        ButtonState lastMouseState = ButtonState.Released;
        // Generatore di numeri casuali usato durante tutto
        // il gioco
        Random gameRandom = new Random();

        // Texture da 1x1 pixel che useremo per disegnare
        // i rettangoli di intersezione
        Texture2D redPixel, greenPixel;
        
        // Costruttore
        public Collisions2DGame()
        {

            // Permettiamo il ridimensionamento della finestra
            this.Window.AllowUserResizing = true;

            // Mostriamo il puntatore
            this.IsMouseVisible = true;

            // Inizializziamo il GraphicsDeviceManager e impostiamo le dimensioni
            // desiderate dell'area di disegno
            graphics = new GraphicsDeviceManager(this);
            graphics.PreferredBackBufferWidth = 800;
            graphics.PreferredBackBufferHeight = 600;

            // Impostiamo la sottocartella con le risorse
            this.Content.RootDirectory = "Content";

        }

        protected override void Initialize()
        {
            // Creiamo un nuovo servizio per il disegno
            spriteBatch = new SpriteBatch(GraphicsDevice);
            this.Services.AddService(typeof(SpriteBatch), spriteBatch);

            base.Initialize();
        }


        // Semplice funzione che crea una Texture 1x1 pixel
        // con colore e trasparenza specificate
        public Texture2D CreatePixel(Color Color, byte Alpha)
        {
            Texture2D px = new Texture2D(this.GraphicsDevice, 1, 1);
            Color.A = Alpha;
            px.SetData<Color>(new Color[] { Color });
            return px;
        }

        protected override void LoadContent()
        {
            // Carichiamo l'immagine dei frame dell'animazione
            framesStellina = this.Content.Load<Texture2D>("stellina");

            // Inizializziamo le texture per disegnare i rettangoli
            // di intersezione
            redPixel = CreatePixel(Color.Red, 127);
            greenPixel = CreatePixel(Color.Lime, 127);
        }

        // Gestiamo la logica di gioco
        protected override void Update(GameTime gameTime)
        {
            if (Keyboard.GetState(PlayerIndex.One).IsKeyDown(Keys.Escape))
                this.Exit();

            // Prendiamo lo stato corrente del mouse
            MouseState mouse = Mouse.GetState();

            // Se  stato rilasciato il mouse
            if (mouse.LeftButton == ButtonState.Released && lastMouseState == ButtonState.Pressed) {

                // Creiamo l'animazione della stellina che si
                // ingrandisce
                Animation cresce = new Animation(framesStellina, 190, 25, 1);
                // Impostiamo una velocit angolare casuale
                cresce.AngularVelocity = (float)gameRandom.NextDouble() * 4 - 2;
                // Impostiamo una dimensione casuale tra le dimensioni
                // originali dell'immagine e la loro met
                cresce.Scale = (float)gameRandom.NextDouble() * 0.5f + 0.5f;

                // Cloniamo l'animazione appena creata e impostiamo
                // che vada al contrario
                Animation decresce = (Animation) ((ICloneable) cresce).Clone();
                decresce.Backward = true;
                
                // Inizializziamo la stellina
                Character stellina = new Character(this, new Vector2(mouse.X, mouse.Y));
                // Impostiamo le animazioni della stellina
                stellina.Animations = new Animation[] { cresce, decresce };

                // Aggiungiamo la stellina appena creata ai
                // componenti del gioco
                this.Components.Add(stellina);
            }

            // Aggiorniamo qual  lo stato del tasto sinistro
            lastMouseState = mouse.LeftButton;

            base.Update(gameTime);
        }

        // Gestiamo l'output grafico
        protected override void Draw(GameTime gameTime)
        {
            // Puliamo l'output grafico
            GraphicsDevice.Clear(Color.SkyBlue);

            // Iniziamo a disegnare
            this.spriteBatch.Begin();

            // Prima di aver disegnato le stelline controlliamo le
            // collisioni effetuando un confronto tra tutti
            // i componenti di tipo Character.
            // Poich la relazione "collide con"  simmetrica
            // effettuiamo solo la met dei controlli.
            for (int c1 = 0; c1 < this.Components.Count - 1; c1++)
                for (int c2 = c1 + 1; c2 < this.Components.Count; c2++)
                    // Ci assicuriamo che si tratti di un Character
                    if (this.Components[c1] is Character && this.Components[c2] is Character)
                    {
                        // Variabile in cui verr salvato il rettangolo di
                        // intersezione tra i due rettangoli che includono
                        // le figure che stiamo confrontando
                        Rectangle intersection;
                        // Effettuiamo un controllo per vedere se c' stata collisione
                        if (Character.CheckCollision((Character)this.Components[c1], 
                            (Character)this.Components[c2], out intersection))
                        {
                            // C' collisione, disegniamo il rettangolo di intersezione
                            // in rosso
                            spriteBatch.Draw(redPixel, (Rectangle)intersection, Color.White);
                        }
#if !DRAW_JUST_COLLISIONS
                        else
                        {
                            // Non c' collisione: le due figure non hanno pixel
                            // non trasparenti sovrapposti.

                            // I rettangoli che contengono le due figure potrebbero
                            // intersecarsi comunque, li disegniamo in verde
                            if (intersection.Width > 0)
                                spriteBatch.Draw(greenPixel, (Rectangle)intersection, Color.White);
                        }
#endif
                    }

            // Chiamiamo il metodo Draw di tutti i componenti
            base.Draw(gameTime);
            // Terminiamo il disegnare
            this.spriteBatch.End();
        }
    }
}
